/* template.c -- example OpenLDAP source file */
/* $OpenLDAP$ */
/* This work is part of OpenLDAP Software <http://www.openldap.org/>.
 *
 * Copyright YEAR The OpenLDAP Foundation.
 * Portions Copyright YEAR Secondary Rights Holder.
 * Portions Copyright YEAR Another Rights Holder.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted only as authorized by the OpenLDAP
 * Public License.
 *
 * A copy of this license is available in the file LICENSE in the
 * top-level directory of the distribution or, alternatively, at
 * <http://www.OpenLDAP.org/license.html>.
 */
/* Additional (custom) notices (where necessary).
 *  Please consult Kurt Zeilenga <kurt@openldap.org> before adding
 *  additional notices.
 */
/* ACKNOWLEDGEMENTS:
 * This work was initially developed by Jane Doe for inclusion in
 * OpenLDAP Software.  Additional significant contributors include:
 *   John Doe
 */
